/* =============================================================================
 UPC - EETAC - CSD - http://digsys.cat

Project P10 tutorial on the use of edge detection interrupts.
 This is a simple FSM that will move when the right flag is activated by
 an edge detection

A LED will turn on when the correcponding input is detected. 

Use only the PORTB
Let's use the INT interrupt (RB0)

Let's use 3 of the 4 RB inputs to detect edges (RB4, BB5 and RB6).

- Interrupt when clicking (falling edge) and then,
  in order to set the flags, in the RB interrupt routine we have to:
    1) Detect the click and wait until the key is released to continue.
    2) Determine which pin was pressed

The other PORTB lines are outputs (RB1, RB2, RB3 and RB7)

Compiler XC8, Microcontroller PIC16F877A
   ========================================================================== */

#include <xc.h>
   /* This general header file includes the device specific headers like
    <pic16f877a.h>. This file contains all the microcontroller
    declarations. All the header files are located in the C compilers's specific
    of: C:\Program Files (x86)\Microchip\xc8\v1.33\include\pic16f877a.h */

/* ------------------------------------
	Definitions
   -------------------------------------------------------------------------- */
/* The enumeration of all the machine states as a "Char" variable
   containing an ASCII character will simplify the debugging.
   "current_state" will contain only a valid state on the list: */

#define 	Idle 		'A'	
#define 	State_INT	'B'
#define 	State_RB6 	'C'
#define 	State_RB5 	'D'
#define 	State_RB4	'E'

/* ------------------------------------
	Function prototypes
   -------------------------------------------------------------------------- */
// The three basic functions to organise the code to emulate a FSM
void init_system (void);         
char state_logic(void);
char output_logic(void);                     

//The interrupt service routine (ISR) to detect the CLK_Flag (external INT0)
// or the ST_SP signal (RB interrupt, or the TMR0 interrupt
static void interrupt service_routine (void); 

// The auxiliary functions to read (poll) and write port pins and be able to
// organise the software variables
void write_outputs (void);
void read_inputs(void);   

/* -----------------------------------
	Global variables
   -------------------------------------------------------------------------- */
/* Let's use convenient byte-type (char) variables to arganise the RAM. */

char current_state = Idle; /* The state variable: a RAM memory position which
                              will be updated ("asynchronously") 
                              every time that "state_logic()" is executed */

char Q_RB_value; // RB(6..4) outputs connected to LED
char Q_INT_value; // The INT output connected to another LED

char INT_Flag;	// The flag that will be set in the ISR
		// to indicate that a new pulse has arrived (INT falling edge)

char RB6_Flag;	// The flag that will be set from the ISR
char RB5_Flag;	// The flag that will be set from the ISR
char RB4_Flag;	// The flag that will be set from the ISR
		
char New_transition_Flag = 0; /* This flag speed the software because
	prevents execution of the output function when there is no change of state */

/* =============================================================================
		Main function
   ========================================================================== */
void main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
	if (state_logic()) break;
	if (output_logic()) break;
    }
    while (1){
        __nop();// the system has crashed in an illegal state and needs attention
                // this loop is never reached in normal operation
    }   
}

/* ------------------------------------
	Function definitions
   -------------------------------------------------------------------------- */

/* =============================================================================
	Init the microcontroller
   ========================================================================== */
void init_system (void){
    char PORTB_value;
 /* These are the PIC16F877A configuration bits translated to the XC8 Compiler.
    They can be generated automatically in the MPLAB X IDE, but normally,
    you don't have to change them.
    So, copy and paste them as they are to your projects */

//#pragma config FOSC = EXTRC   // Oscillator Selection bits (RC oscillator)
#pragma config FOSC = XT     	// External XTAL Oscillator

//#pragma config WDTE = ON      // Watchdog Timer Enable bit (WDT enabled)
#pragma config WDTE = OFF       // Watchdog Timer not in use (WDT disabled)
//#pragma config PWRTE = OFF    // Power-up Timer Enable bit (PWRT disabled)

#pragma config PWRTE = ON      	// Power-up Timer Enable bit (PWRT enabled)

#pragma config BOREN = ON       // Brown-out Reset Enable bit (BOR enabled)
#pragma config LVP = ON         // Low-Voltage (Single-Supply) In-Circuit Serial
				// Programming Enable bit (RB3/PGM pin has PGM
				// function; low-voltage programming enabled)

#pragma config CPD = OFF        // Data EEPROM Memory Code Protection bit
        			//(Data EEPROM code protection off)
#pragma config WRT = OFF        // Flash Program Memory Write Enable bits (Write
				// protection off; all program memory may be
				// written to by EECON control)
#pragma config CP = OFF         // Flash Program Memory Code Protection bit
                		// (Code protection off)

// Inicialise operations starts here:
    /* Pay attention to how the ports are configured
       using mainly PORTx and TRISx registers*/
    PORTB 	= 0x00; 	// Reset all Flip-Flops at PORTB
    TRISB 	= 0b01110001; /* PORTB pin directions */

    INTEDG	= 0;	// INT edge selection: '0' --> on falling edge
                        // (interrupt after clicking the pushbutton)

    SSPEN = 0;          // I2C bus disabled
    nRBPU  = 0;// RB port Pull-up resistors enabled to spare exernal resistors

// Let's init the state machine:
    current_state = Idle;
    INT_Flag = 0;
    RB6_Flag = 0;
    RB5_Flag = 0;
    RB4_Flag = 0;

    // Let's init the counter outputs:
    Q_RB_value = 0b00000000;
    Q_INT_value = 0;

        /* This flag will allow to speed up the code: outputs will be written
       only when there is a change */
    New_transition_Flag = 0;
    write_outputs();

    INTF   = 0;
    INTE   = 1;	// Enable INT interrupts
    RBIE   = 1;	// Enable RB interrupts
    PORTB_value = PORTB;
    RBIF   = 0;
    
    GIE = 1;	// Global interrupts allowed
}

//*****************************************************************************
//State variable logic routine 
//*****************************************************************************
char state_logic(void){
    // local variable to check tha the FSM is always in a known state
    char error = 0;
    
    read_inputs();  // Sample the level sensitive inputs

    switch (current_state) {
        case Idle:
            if (INT_Flag == 1){
                current_state = State_INT;
                New_transition_Flag = 1; // jump to the next state, and so, 
                                	// write the new output values
                INT_Flag = 0; INT_Flag = 0; RB6_Flag= 0;RB5_Flag= 0;RB4_Flag= 0;
            }
            else // The loop, only for making the code more readible    
                current_state = Idle;	
	    break;

        case State_INT:
            if (RB6_Flag){
                current_state = State_RB6;
                New_transition_Flag = 1;
                RB6_Flag = 0;
            }
        break;
        
        case State_RB6: 
            if (RB5_Flag){
                current_state = State_RB5;
                New_transition_Flag = 1;
                RB5_Flag = 0;

            }
       
        break;

        case State_RB5:
            if (RB4_Flag){
                current_state = State_RB4;
                New_transition_Flag = 1;
                RB4_Flag= 0;
            }
               
            break;
        case State_RB4:
            if (INT_Flag){
                current_state = Idle;
                New_transition_Flag = 1;
                // let's clear all the flags because it goes home
                INT_Flag = 0; RB6_Flag= 0;RB5_Flag= 0;RB4_Flag= 0;
            }
            break;

        default:
            error = 1; // If no states selected, there is an error
                       // and the systems must be stopped
            break;
    }    
  
    return (error);
}

//*****************************************************************************
//Output logic routine 
//*****************************************************************************
char output_logic(void){
    char error = 0;
// Re-write outputs when there is a change of state only
    if (New_transition_Flag == 1){
	    switch (current_state) { // The CC2 truth table
	        case Idle:
	            Q_RB_value = 0b00000000;
	            Q_INT_value = 0;
                    break;
	
  	        case State_INT:
	            Q_RB_value = 0b00000000;
	            Q_INT_value = 1;
	            break;

	        case State_RB6:
	            Q_RB_value = 0b00000100;
	            Q_INT_value = 0;
	
	            break;

	        case State_RB5:
	            Q_RB_value = 0b00000010;
	            Q_INT_value = 0;
	
	            break;

	        case State_RB4:
	            Q_RB_value = 0b00000001;
	            Q_INT_value = 0;
	
	            break;
	            	            	            
	        default:             
	            Q_RB_value = 0b00000000;
	            Q_INT_value = 0;	        
	            error = 1;
                        // If no states have been selected, there is an error
                        // and the systems must be stopped
	            break;
	}
    write_outputs();
    New_transition_Flag = 0; // So, the output RAM variable can bw written again
}

    return (error);
}

/* **************************************************************************
   ISR: Interrupt Service Routine to attend the external INT
   "interrupt" is the key word here to make this ISR function special
   ***************************************************************************/
static void interrupt service_routine (void){
    char  i, k, PORTB_read;
    GIE = 0;    //Disable interrupts while attending one of them ...

    if(INTF) { // INT interrupt? (if so, the hardware flag bit is set)
              // Read the PIC16F877A datasheet, Chapter 14, section 11.
        INT_Flag = 1;   // Let's set our convenient "software" flag,
              // so we'll update the state every time that there is an interrupt
        INTF = 0;       // Clear interrupt "hardware" flag
                     // So that the system can be interrupted again
        RBIE = 1; //Enable now the RB interrupts
    }
    if(RBIF) { // RB interrupt? (if so, the hardware flag bit is set)
        PORTB_read = PORTB & 0b01110000; /*save the PORTB value to determine
                                          which key is pressed after releasing*/
        i = 0;
        while (i <5 ){  // wait until the key is debounced and released
                        // the delay is programmable using the i and k
            if ((PORTB & 0b01110000) == 0x70) i ++;
            else i =0;
            for (k=0; k < 10;k++){;} //delay
        };
    // and now, once the key is released, let's determine which one was clicked:
        if (PORTB_read == 0x30) RB6_Flag=1;
        else if (PORTB_read == 0x50) RB5_Flag=1;
        else if (PORTB_read == 0x60) RB4_Flag=1;
                
        RBIF = 0;  // the PORTB is stable now, an so, we can reset the flag
    }

    GIE = 1;            //Enable general interrupts
}

/* =============================================================================
 Scan inputs and process the bits so that we can set our convenient variables.
 Many of them are "char" to be monitored easily using the watch window.
  =========================================================================== */
void read_inputs (void){

}

/* =============================================================================
	Driving the output pins. Let's drive the output ports, converting
	microcontroller variables into electrical signals.
   ========================================================================== */
  void write_outputs(void){
	char PORTB_value = 0;// local variable to prepare the data to be written 
	
	PORTB_value =  (Q_INT_value <<7) | (Q_RB_value <<1);
	PORTB = PORTB_value ^0xFF; // All LED active low
	
	/* Pay attention that this operation is not possible here: 
	Do not read the PORTB when using RB interrupts:
        PORTB = ( (PORTB & 0b10001110) | (Q_INT_value <<7) | (Q_RB_value <<1)*/

}


